
// sequence of elements in S[1..n] is sorted in a way that is compatible with the partial order defined by Graph
function compatible_sort (Graph: [int, int] bool, S: [int] int, n: int) returns (bool);

// definition for base case
axiom ( forall Graph: [int, int] bool, S: [int] int, n: int :: n == 0  ==>  compatible_sort(Graph, S, n) );
// definition for inductive step
axiom ( forall Graph: [int, int] bool, S: [int] int, n: int :: n > 0 && compatible_sort(Graph, S, n-1) &&
	                                                             ( forall a: int :: Graph[a, S[n]] ==> (exists b: int :: 1 <= b && b < n && S[b] == a) )
																   ==>  compatible_sort(Graph, S, n) );
// this is needed to lift compatible_sort(DAG, S, n) after the assignement S[n] := x
axiom ( forall Graph: [int, int] bool, S: [int] int, T: [int] int, n: int :: n > 0 && compatible_sort(Graph, T, n) && ( forall i: int :: 1 <= i && i <= n ==> T[i] == S[i] )
																				==>  compatible_sort(Graph, S, n) );

																				
					 // DAG[x, y] iff there is an edge from node x to node y
					 // E is a set of elements (represented by its indicator function)
procedure topo_sort (DAG: [int, int] bool, E: [int] bool)
			// list S of n elements
	returns ( S: [int] int, n: int )
	
	// every element in S is from E
	ensures ( forall i: int :: 1 <= i && i <= n  ==>  E[S[i]] );
	// every element in E is also somewhere in S
	ensures ( forall i: int :: E[i]  ==>  (exists j: int :: 1 <= j && j <= n && S[j] == i) );
	// the sequence S is compatible with the partial order defined by DAG
	ensures ( compatible_sort(DAG, S, n) );
{
	var x: int;
	var G, H: [int, int] bool;
	var elements: [int] bool;
	var exit_now: bool;
	
	G := DAG;
	elements := E;
	n := 0;

	while ( (exists i: int :: elements[i]) )
	
	invariant ( forall i: int :: elements[i] ==> E[i] );
	// invariant ( forall i: int :: 1 <= i && i <= n  ==>  E[S[i]] );
	
	invariant ( forall a: int, b: int ::  G[a,b] ==> DAG[a,b] );
	invariant ( forall a: int, b: int :: DAG[a,b] && !G[a,b] ==> ( exists c: int :: 1 <= c && c <= n && S[c] == a ) );
	invariant ( n > 0  ==>  (forall a: int :: DAG[a, S[n]] ==> (exists b: int :: 1 <= b && b < n && S[b] == a) ));
	// invariant ( compatible_sort(DAG, S, n) );
	// invariant ( forall i: int :: ! elements[i] && E[i]  ==>  (exists j: int :: 1 <= j && j <= n && S[j] == i) );
	{
		// let x be an element without predecessors for the constraints
		havoc x;
		assume ( forall a: int :: ! G[a, x] );
		assume ( elements[x] );
		// produce x as the next element of S 
		n := n + 1;
		S[n] := x;
		// remove x from the set of elements
		elements[x] := false;
		// remove all pairs starting with x from the set of constraints
		havoc H;
		assume ( forall b: int :: ! H[x, b] );
		assume ( forall a: int, b: int :: H[a, b] ==> G[a, b] );
		assume ( forall a: int, b: int :: x != a && G[a, b] ==> H[a, b] );
		G := H;
	}
}	
